/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Assorted globals and setup/cleanup routines.
 *
 *      By Shawn Hargreaves.
 *
 *      See readme.txt for copyright information.
 */


#include <stdio.h>
#include <ctype.h>

#include "allegro.h"
#include "allegro/aintern.h"



/* in case you want to report version numbers */
char allegro_id[] = "Allegro " ALLEGRO_VERSION_STR;


/* error message for sound and gfx init routines */
char allegro_error[256] = EMPTY_STRING;


/* current system driver */
SYSTEM_DRIVER *system_driver = NULL;


/* flag for how many times we have been initialised */
int _allegro_count = 0;


/* the graphics driver currently in use */
GFX_DRIVER *gfx_driver = NULL;


/* abilities of the current graphics driver */
int gfx_capabilities = 0;


/* a bitmap structure for accessing the physical screen */
BITMAP *screen = NULL;


/* info about the current graphics drawing mode */
int _drawing_mode = DRAW_MODE_SOLID;

BITMAP *_drawing_pattern = NULL;

int _drawing_x_anchor = 0;
int _drawing_y_anchor = 0;

unsigned int _drawing_x_mask = 0;
unsigned int _drawing_y_mask = 0;


/* default palette structures */
PALETTE black_palette;
PALETTE _current_palette; 

RGB black_rgb = { 0, 0, 0 };

PALETTE desktop_palette = 
{
   { 63, 63, 63 },   { 63, 0,  0  },
   { 0,  63, 0  },   { 63, 63, 0  },
   { 0,  0,  63 },   { 63, 0,  63 },
   { 0,  63, 63 },   { 16, 16, 16 },
   { 31, 31, 31 },   { 63, 31, 31 },
   { 31, 63, 31 },   { 63, 63, 31 },
   { 31, 31, 63 },   { 63, 31, 63 },
   { 31, 63, 63 },   { 0,  0,  0  }
};


/* a block of temporary working memory */
void *_scratch_mem = NULL;
int _scratch_mem_size = 0;


/* SVGA bank switching tables */
int _last_bank_1 = -1;
int _last_bank_2 = -1;
int *_gfx_bank = NULL;


/* what OS are we running under? */
int os_type = OSTYPE_UNKNOWN;


/* vertical retrace counter and callback function */
volatile int retrace_count = 0; 
void (*retrace_proc)(void) = NULL;


/* processor information */
char cpu_vendor[32] = EMPTY_STRING;
int cpu_family = 0;
int cpu_model = 0;
int cpu_fpu = FALSE; 
int cpu_mmx = FALSE; 
int cpu_3dnow = FALSE;
int cpu_cpuid = FALSE; 


/* debugging stuff */
static int debug_virgin = TRUE;

static FILE *assert_file = NULL;
static FILE *trace_file = NULL;

static int (*assert_handler)(char *msg) = NULL;
static int (*trace_handler)(char *msg) = NULL;


/* dynamic registration system for cleanup code */
#define MAX_EXIT_FUNCS     16

typedef void (*FUNCPTR)(void);

static FUNCPTR exit_funcs[MAX_EXIT_FUNCS] = 
{ 
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
   NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};



/* _add_exit_func:
 *  Adds a function to the list that need to be called by allegro_exit().
 */
void _add_exit_func(void (*func)(void))
{
   int c;

   for (c=0; c<MAX_EXIT_FUNCS; c++) {
      if ((!exit_funcs[c]) || (exit_funcs[c] == func)) {
	 exit_funcs[c] = func;
	 return;
      }
   }
}



/* _remove_exit_func:
 *  Removes a function from the list that need to be called by allegro_exit().
 */
void _remove_exit_func(void (*func)(void))
{
   int c;

   for (c=0; c<MAX_EXIT_FUNCS; c++) {
      if (exit_funcs[c] == func) {
	 while (c<MAX_EXIT_FUNCS-1) {
	    exit_funcs[c] = exit_funcs[c+1];
	    c++;
	 }
	 exit_funcs[MAX_EXIT_FUNCS-1] = NULL;
	 break;
      }
   }
}



/* allegro_init:
 *  Initialises the Allegro library, activating the system driver.
 */
int install_allegro(int system_id)
{
   int i;

   #ifndef CONSTRUCTOR_FUNCTION
      /* call constructor functions manually */
      extern void _initialize_datafile_types();
      extern void _midi_constructor();

      _initialize_datafile_types();
      _midi_constructor();
   #endif

   /* set up default palette structures */
   for (i=0; i<256; i++)
      black_palette[i] = black_rgb;

   for (i=16; i<256; i++)
      desktop_palette[i] = desktop_palette[i & 15];

   /* initialise the system driver */
   for (i=0; _system_driver_list[i].driver; i++) {
      if ((_system_driver_list[i].id == system_id) ||
	  ((_system_driver_list[i].autodetect) && (system_id == SYSTEM_AUTODETECT))) {
	 system_driver = _system_driver_list[i].driver;
	 if (system_driver->init() != 0) {
	    system_driver = NULL;
	    if (system_id != SYSTEM_AUTODETECT)
	       break;
	 }
      }
   }

   if (!system_driver)
      return -1;

   /* sort out text translations and CPU type */
   _load_config_text();

   system_driver->name = system_driver->desc = get_config_text(system_driver->ascii_name);

   check_cpu();

   /* install shutdown handler */
   if (_allegro_count == 0)
      atexit(allegro_exit);

   _allegro_count++;

   return 0;
}



/* allegro_exit:
 *  Closes down the Allegro system.
 */
void allegro_exit()
{
   int done, c;

   do {
      done = TRUE;

      for (c=MAX_EXIT_FUNCS-1; c>=0; c--) {
	 if (exit_funcs[c]) {
	    (*exit_funcs[c])();
	    done = FALSE;
	 }
      }
   } while (!done);

   if (system_driver) {
      system_driver->exit();
      system_driver = NULL;
   }

   if (_scratch_mem) {
      free(_scratch_mem);
      _scratch_mem = NULL;
      _scratch_mem_size = 0;
   }
}



/* debug_exit:
 *  Closes the debugging output files.
 */
static void debug_exit(void)
{
   if (assert_file) {
      fclose(assert_file);
      assert_file = NULL;
   }

   if (trace_file) {
      fclose(trace_file);
      trace_file = NULL;
   }

   _remove_exit_func(debug_exit);
}



/* debug_init:
 *  Opens the debugging output files.
 */
static void debug_init()
{
   char *s;

   s = getenv("ALLEGRO_ASSERT");

   if (s)
      assert_file = fopen(s, "w");
   else
      assert_file = NULL;

   s = getenv("ALLEGRO_TRACE");

   if (s)
      trace_file = fopen(s, "w");
   else
      trace_file = fopen("allegro.log", "w");

   _add_exit_func(debug_exit);

   debug_virgin = FALSE;
}



/* al_assert:
 *  Raises an assert.
 */
void al_assert(char *file, int line)
{
   static int asserted = FALSE;
   char buf[80];

   if (asserted)
      return;

   sprintf(buf, "Assert failed at line %d of %s", line, file);

   if (assert_handler) {
      if (assert_handler(buf))
	 return;
   }

   if (debug_virgin)
      debug_init();

   if (assert_file) {
      fprintf(assert_file, "%s\n", buf);
      fflush(assert_file);
   }
   else {
      asserted = TRUE;

      if ((system_driver) && (system_driver->assert)) {
	 system_driver->assert(buf);
      }
      else {
	 allegro_exit();
	 printf("%s\n", buf);
	 exit(1);
      }
   }
}



/* al_trace:
 *  Outputs a trace message.
 */
void al_trace(char *msg, ...)
{
   char buf[256];

   va_list ap;
   va_start(ap, msg);
   uvsprintf(buf, msg, ap);
   va_end(ap);

   if (trace_handler) {
      if (trace_handler(msg))
	 return;
   }

   if (debug_virgin)
      debug_init();

   if (trace_file) {
      fwrite(buf, 1, ustrsize(buf), trace_file);
      fflush(trace_file);
   }
}



/* register_assert_handler:
 *  Installs a user handler for assert failures.
 */
void register_assert_handler(int (*handler)(char *msg))
{
   assert_handler = handler;
}



/* register_trace_handler:
 *  Installs a user handler for trace output.
 */
void register_trace_handler(int (*handler)(char *msg))
{
   trace_handler = handler;
}



#ifdef ALLEGRO_NO_STRICMP


/* _al_stricmp:
 *  Replacement routine for compilers that have no stricmp() library function.
 */
int _al_stricmp(char *s1, char *s2)
{
   int c1, c2;

   for (;;) {
      c1 = *(s1++);
      c2 = *(s2++);

      c1 = tolower(c1);
      c2 = tolower(c2);

      if (c1 != c2)
	 return c1 - c2;

      if (!c1)
	 return 0;
   }
}


#endif


