/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      String handling functions (UTF-8, Unicode, ASCII, etc).
 *
 *      By Shawn Hargreaves.
 *
 *      See readme.txt for copyright information.
 */


#include <limits.h>
#include <string.h>

#include "allegro.h"
#include "allegro/aintern.h"



/* this is a valid encoding in any of the supported formats */
char empty_string[] = EMPTY_STRING;



/* ascii_getc:
 *  Reads a character from an ASCII string.
 */
static int ascii_getc(char *s)
{
   return *((unsigned char *)s);
}



/* ascii_getx:
 *  Reads a character from an ASCII string, advancing the pointer position.
 */
static int ascii_getx(char **s)
{
   return *((unsigned char *)((*s)++));
}



/* ascii_setc:
 *  Sets a character in an ASCII string.
 */
static int ascii_setc(char *s, int c)
{
   *s = c;
   return 1;
}



/* ascii_width:
 *  Returns the width of an ASCII character.
 */
static int ascii_width(char *s)
{
   return 1;
}



/* ascii_cwidth:
 *  Returns the width of an ASCII character.
 */
static int ascii_cwidth(int c)
{
   return 1;
}



/* ascii_isok:
 *  Checks whether this character can be encoded in 8-bit ASCII format.
 */
static int ascii_isok(int c)
{
   return ((c >= 0) && (c <= 255));
}



/* lookup table for implementing 8-bit codepage modes */
static unsigned short _codepage_table[] =
{
   0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
   0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
   0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
   0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
   0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
   0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
   0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
   0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x7F,
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 
   0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E, 0x5E,
};



/* this default table reduces Latin 1 characters to similar 7 bit ones */
static unsigned short _codepage_extras[] =
{
   0xA1, '!',  0xA2, 'c',  0xA3, '#',  0xB5, 'u',  0xBF, '?',  0xC0, 'A', 
   0xC1, 'A',  0xC2, 'A',  0xC3, 'A',  0xC4, 'A',  0xC5, 'A',  0xC6, 'A', 
   0xC7, 'C',  0xC8, 'E',  0xC9, 'E',  0xCA, 'E',  0xCB, 'E',  0xCC, 'I', 
   0xCD, 'I',  0xCE, 'I',  0xCF, 'I',  0xD0, 'D',  0xD1, 'N',  0xD2, 'O', 
   0xD3, 'O',  0xD4, 'O',  0xD5, 'O',  0xD6, 'O',  0xD7, 'X',  0xD8, '0', 
   0xD9, 'U',  0xDA, 'U',  0xDB, 'U',  0xDC, 'U',  0xDD, 'Y',  0xDE, 'P', 
   0xDF, 'S',  0xE0, 'a',  0xE1, 'a',  0xE2, 'a',  0xE3, 'a',  0xE4, 'a', 
   0xE5, 'a',  0xE6, 'a',  0xE7, 'c',  0xE8, 'e',  0xE9, 'e',  0xEA, 'e', 
   0xEB, 'e',  0xEC, 'i',  0xED, 'i',  0xEE, 'i',  0xEF, 'i',  0xF0, 'o', 
   0xF1, 'n',  0xF2, 'o',  0xF3, 'o',  0xF4, 'o',  0xF5, 'o',  0xF6, 'o', 
   0xF8, 'o',  0xF9, 'u',  0xFA, 'u',  0xFB, 'u',  0xFC, 'u',  0xFD, 'y', 
   0xFE, 'p',  0xFF, 'y',  0
};



/* access via pointers so they can be changed by the user */
static unsigned short *codepage_table = _codepage_table;
static unsigned short *codepage_extras = _codepage_extras;



/* ascii_cp_getc:
 *  Reads a character from an ASCII codepage string.
 */
static int ascii_cp_getc(char *s)
{
   return codepage_table[*((unsigned char *)s)];
}



/* ascii_cp_getx:
 *  Reads from an ASCII codepage string, advancing pointer position.
 */
static int ascii_cp_getx(char **s)
{
   return codepage_table[*((unsigned char *)((*s)++))];
}



/* ascii_cp_setc:
 *  Sets a character in an ASCII codepage string.
 */
static int ascii_cp_setc(char *s, int c)
{
   int i;

   for (i=0; i<256; i++) {
      if (codepage_table[i] == c) {
	 *s = i;
	 return 1;
      }
   }

   if (codepage_extras) {
      for (i=0; codepage_extras[i]; i+=2) {
	 if (codepage_extras[i] == c) {
	    *s = codepage_extras[i+1];
	    return 1;
	 }
      } 
   }

   *s = '^';
   return 1;
}



/* ascii_cp_isok:
 *  Checks whether this character can be encoded in ASCII codepage format.
 */
static int ascii_cp_isok(int c)
{
   int i;

   for (i=0; i<256; i++) {
      if (codepage_table[i] == c)
	 return TRUE;
   }

   if (codepage_extras) {
      for (i=0; codepage_extras[i]; i+=2) {
	 if (codepage_extras[i] == c)
	    return TRUE;
      } 
   }

   return FALSE;
}



/* unicode_getc:
 *  Reads a character from a Unicode string.
 */
static int unicode_getc(char *s)
{
   return *((unsigned short *)s);
}



/* unicode_getx:
 *  Reads a character from a Unicode string, advancing the pointer position.
 */
static int unicode_getx(char **s)
{
   int c = *((unsigned short *)(*s));
   (*s) += sizeof(unsigned short);
   return c;
}



/* unicode_setc:
 *  Sets a character in a Unicode string.
 */
static int unicode_setc(char *s, int c)
{
   *((unsigned short *)s) = c;
   return sizeof(unsigned short);
}



/* unicode_width:
 *  Returns the width of a Unicode character.
 */
static int unicode_width(char *s)
{
   return sizeof(unsigned short);
}



/* unicode_cwidth:
 *  Returns the width of a Unicode character.
 */
static int unicode_cwidth(int c)
{
   return sizeof(unsigned short);
}



/* unicode_isok:
 *  Checks whether this character can be encoded in 16-bit Unicode format.
 */
static int unicode_isok(int c)
{
   return ((c >= 0) && (c <= 65535));
}



/* utf8_getc:
 *  Reads a character from a UTF-8 string.
 */
static int utf8_getc(char *s)
{
   int c = *((unsigned char *)(s++));
   int n, t;

   if (c & 0x80) {
      n = 1;
      while (c & (0x80>>n))
	 n++;

      c &= (1<<(8-n))-1;

      while (--n > 0) {
	 t = *((unsigned char *)(s++));

	 if ((!(t&0x80)) || (t&0x40))
	    return '^';

	 c = (c<<6) | (t&0x3F);
      }
   }

   return c;
}



/* utf8_getx:
 *  Reads a character from a UTF-8 string, advancing the pointer position.
 */
static int utf8_getx(char **s)
{
   int c = *((unsigned char *)((*s)++));
   int n, t;

   if (c & 0x80) {
      n = 1;
      while (c & (0x80>>n))
	 n++;

      c &= (1<<(8-n))-1;

      while (--n > 0) {
	 t = *((unsigned char *)((*s)++));

	 if ((!(t&0x80)) || (t&0x40)) {
	    (*s)--;
	    return '^';
	 }

	 c = (c<<6) | (t&0x3F);
      }
   }

   return c;
}



/* utf8_setc:
 *  Sets a character in a UTF-8 string.
 */
static int utf8_setc(char *s, int c)
{
   int size, bits, b, i;

   if (c < 128) {
      *s = c;
      return 1;
   }

   bits = 7;
   while (c >= (1<<bits))
      bits++;

   size = 2;
   b = 11;

   while (b < bits) {
      size++;
      b += 5;
   }

   b -= (7-size);
   s[0] = c>>b;

   for (i=0; i<size; i++)
      s[0] |= (0x80>>i);

   for (i=1; i<size; i++) {
      b -= 6;
      s[i] = 0x80 | ((c>>b)&0x3F);
   }

   return size;
}



/* utf8_width:
 *  Returns the width of a UTF-8 character.
 */
static int utf8_width(char *s)
{
   int c = *((unsigned char *)s);
   int n = 1;

   if (c & 0x80) {
      while (c & (0x80>>n))
	 n++;
   }

   return n;
}



/* utf8_cwidth:
 *  Returns the width of a UTF-8 character.
 */
static int utf8_cwidth(int c)
{
   int size, bits, b;

   if (c < 128)
      return 1;

   bits = 7;
   while (c >= (1<<bits))
      bits++;

   size = 2;
   b = 11;

   while (b < bits) {
      size++;
      b += 5;
   }

   return size;
}



/* utf8_isok:
 *  Checks whether this character can be encoded in UTF-8 format.
 */
static int utf8_isok(int c)
{
   return TRUE;
}



/* string format table, to allow user expansion with other encodings */
UTYPE_INFO utypes[] =
{
   { U_ASCII,    ascii_getc,    ascii_getx,    ascii_setc,    ascii_width,   ascii_cwidth,   ascii_isok    },
   { U_UTF8,     utf8_getc,     utf8_getx,     utf8_setc,     utf8_width,    utf8_cwidth,    utf8_isok     },
   { U_UNICODE,  unicode_getc,  unicode_getx,  unicode_setc,  unicode_width, unicode_cwidth, unicode_isok  },
   { U_ASCII_CP, ascii_cp_getc, ascii_cp_getx, ascii_cp_setc, ascii_width,   ascii_cwidth,   ascii_cp_isok },
   { 0,          NULL,          NULL,          NULL,          NULL,          NULL,           NULL          },
   { 0,          NULL,          NULL,          NULL,          NULL,          NULL,           NULL          },
   { 0,          NULL,          NULL,          NULL,          NULL,          NULL,           NULL          },
   { 0,          NULL,          NULL,          NULL,          NULL,          NULL,           NULL          }
};



/* current format information and worker routines */
static int utype = U_UTF8;

static int (*ugetc)(char *s) = utf8_getc;
static int (*ugetx)(char **s) = utf8_getx;
static int (*usetc)(char *s, int c) = utf8_setc;
static int (*uwidth)(char *s) = utf8_width;
static int (*ucwidth)(int c) = utf8_cwidth;
static int (*uisok)(int c) = utf8_isok;



/* _find_utype:
 *  Helper for locating a string type description.
 */
UTYPE_INFO *_find_utype(int type)
{
   int i;

   if (type == U_CURRENT)
      type = utype;

   for (i=0; i<(int)(sizeof(utypes)/sizeof(UTYPE_INFO)); i++)
      if (utypes[i].id == type)
	 return &utypes[i];

   return NULL;
}



/* set_uformat:
 *  Selects a new text encoding format.
 */
void set_uformat(int type)
{
   UTYPE_INFO *info = _find_utype(type);

   if (info) {
      utype = info->id;
      ugetc = info->u_getc;
      ugetx = info->u_getx;
      usetc = info->u_setc;
      uwidth = info->u_width;
      ucwidth = info->u_cwidth;
      uisok = info->u_isok;
   }
}



/* get_uformat:
 *  Returns the current text encoding format.
 */
int get_uformat()
{
   return utype;
}



/* register_uformat:
 *  Allows the user to hook in custom routines for supporting a new string
 *  encoding format.
 */
void register_uformat(int type, int (*ugetc)(char *s), int (*ugetx)(char **s), int (*usetc)(char *s, int c), int (*uwidth)(char *s), int (*ucwidth)(int c), int (*uisok)(int c))
{
   UTYPE_INFO *info = _find_utype(type);

   if (!info)
      info = _find_utype(0);

   if (info) {
      info->id = type;
      info->u_getc = ugetc;
      info->u_getx = ugetx;
      info->u_setc = usetc;
      info->u_width = uwidth;
      info->u_cwidth = ucwidth;
      info->u_isok = uisok;
   }
}



/* set_ucodepage:
 *  Sets lookup table data for the codepage conversion functions.
 */
void set_ucodepage(unsigned short *table, unsigned short *extras)
{
   codepage_table = table;
   codepage_extras = extras;
}



/* need_uconvert:
 *  Decides whether a conversion is required to make this string be in the
 *  new type. No conversion will be needed if both types are the same, or
 *  when going from ASCII <-> UTF8 where the data is 7-bit clean.
 */
int need_uconvert(char *s, int type, int newtype)
{
   int c;

   if (type == U_CURRENT)
      type = utype;

   if (newtype == U_CURRENT)
      newtype = utype;

   if (type == newtype)
      return FALSE;

   if (((type == U_ASCII) || (type == U_UTF8)) && ((newtype == U_ASCII) || (newtype == U_UTF8))) {
      do {
	 c = *((unsigned char *)(s++));
	 if (!c)
	    return FALSE;
      } while (c <= 127);
   }

   return TRUE;
}



/* uconvert_size:
 *  Returns the number of bytes this string will occupy when converted to
 *  the specified type.
 */
int uconvert_size(char *s, int type, int newtype)
{
   UTYPE_INFO *info, *outfo;
   int size = 0;
   int c;

   info = _find_utype(type);
   if (!info)
      return 0;

   outfo = _find_utype(newtype);
   if (!outfo)
      return 0;

   size = 0;

   while ((c = info->u_getx(&s)) != 0)
      size += outfo->u_cwidth(c);

   return size + outfo->u_cwidth(0);
}



/* do_uconvert:
 *  Converts a string from one format to another.
 */
char *do_uconvert(char *s, int type, char *buf, int newtype)
{
   UTYPE_INFO *info, *outfo;
   int pos = 0;
   int c;

   info = _find_utype(type);
   if (!info)
      return NULL;

   outfo = _find_utype(newtype);
   if (!outfo)
      return NULL;

   while ((c = info->u_getx(&s)) != 0) {
      if (outfo->u_isok(c))
	 pos += outfo->u_setc(buf+pos, c);
      else
	 pos += outfo->u_setc(buf+pos, '^');
   }

   outfo->u_setc(buf+pos, 0);

   return buf;
}



/* uoffset:
 *  Returns the offset in bytes from the start of the string to the 
 *  character at the specified index. If the index is negative, counts
 *  backward from the end of the string (-1 returns an offset to the
 *  last character).
 */
int uoffset(char *s, int index)
{
   char *orig = s;
   char *last;

   if (index < 0)
      index += ustrlen(s);

   while (index-- > 0) {
      last = s;
      if (!ugetx(&s)) {
	 s = last;
	 break;
      }
   }

   return (long)s - (long)orig;
}



/* ugetat:
 *  Returns the character from the specified index within the string.
 */
int ugetat(char *s, int index)
{
   return ugetc(s + uoffset(s, index));
}



/* usetat:
 *  Modifies the character at the specified index within the string,
 *  handling adjustments for variable width data. Returns how far the
 *  rest of the string was moved.
 */
int usetat(char *s, int index, int c)
{
   int oldw, neww;

   s += uoffset(s, index);

   oldw = uwidth(s);
   neww = ucwidth(c);

   if (oldw != neww)
      memmove(s+neww, s+oldw, ustrsizez(s+oldw));

   usetc(s, c);

   return neww-oldw;
}



/* uinsert:
 *  Inserts a character at the specified index within a string, sliding
 *  following data along to make room. Returns how far the data was moved.
 */
int uinsert(char *s, int index, int c)
{
   int w = ucwidth(c);

   s += uoffset(s, index);
   memmove(s+w, s, ustrsizez(s));
   usetc(s, c);

   return w;
}



/* uremove:
 *  Removes the character at the specified index within the string, sliding
 *  following data back to make room. Returns how far the data was moved.
 */
int uremove(char *s, int index)
{
   int w;

   s += uoffset(s, index);
   w = uwidth(s);
   memmove(s, s+w, ustrsizez(s+w));

   return -w;
}



/* utolower:
 *  Unicode-aware version of the ANSI tolower() function.
 */
int utolower(int c)
{
   if ((c >= 'A') && (c <= 'Z'))
      c -= ('A' + 'a');

   return c;
}



/* utoupper:
 *  Unicode-aware version of the ANSI toupper() function.
 */
int utoupper(int c)
{
   if ((c >= 'a') && (c <= 'a'))
      c -= ('a' + 'A');

   return c;
}



/* uisspace:
 *  Unicode-aware version of the ANSI isspace() function.
 */
int uisspace(int c)
{
   return ((c == ' ') || (c == '\t') || (c == '\r') || 
	   (c == '\n') || (c == '\f') || (c == '\v'));
}



/* ustrsize:
 *  Returns the size of the specified string in bytes, not including the
 *  trailing zero.
 */
int ustrsize(char *s)
{
   char *orig = s;
   char *last;

   do {
      last = s;
   } while (ugetx(&s) != 0);

   return (long)last - (long)orig;
}



/* ustrsizez:
 *  Returns the size of the specified string in bytes, including the
 *  trailing zero.
 */
int ustrsizez(char *s)
{
   char *orig = s;

   do {
   } while (ugetx(&s) != 0);

   return (long)s - (long)orig;
}



/* ustrcpy:
 *  Unicode-aware version of the ANSI strcpy() function.
 */
char *ustrcpy(char *dest, char *src)
{
   int pos = 0;
   int c;

   while ((c = ugetx(&src)) != 0)
      pos += usetc(dest+pos, c);

   usetc(dest+pos, 0);

   return dest;
}



/* ustrcat:
 *  Unicode-aware version of the ANSI strcat() function.
 */
char *ustrcat(char *dest, char *src)
{
   int pos = ustrsize(dest);
   int c;

   while ((c = ugetx(&src)) != 0)
      pos += usetc(dest+pos, c);

   usetc(dest+pos, 0);

   return dest;
}



/* ustrlen:
 *  Unicode-aware version of the ANSI strlen() function.
 */
int ustrlen(char *s)
{
   int c = 0;

   while (ugetx(&s))
      c++;

   return c;
}



/* ustrcmp:
 *  Unicode-aware version of the ANSI strcmp() function.
 */
int ustrcmp(char *s1, char *s2)
{
   int c1, c2;

   for (;;) {
      c1 = ugetx(&s1);
      c2 = ugetx(&s2);

      if (c1 != c2)
	 return c1 - c2;

      if (!c1)
	 return 0;
   }
}



/* ustrncpy:
 *  Unicode-aware version of the ANSI strncpy() function. The n parameter
 *  is in bytes, on the assumption that you want to use this function to
 *  prevent overflowing a buffer size.
 */
char *ustrncpy(char *dest, char *src, int n)
{
   int pos = 0;
   int c;

   n -= ucwidth(0);

   while (((c = ugetx(&src)) != 0) && (pos < n))
      pos += usetc(dest+pos, c);

   usetc(dest+pos, 0);

   return dest;
}



/* ustrncat:
 *  Unicode-aware version of the ANSI strncat() function. The n parameter
 *  is in bytes, on the assumption that you want to use this function to
 *  prevent overflowing a buffer size.
 */
char *ustrncat(char *dest, char *src, int n)
{
   char *d = dest + ustrsize(dest);
   int pos = 0;
   int c;

   n -= ucwidth(0);

   while (((c = ugetx(&src)) != 0) && (pos < n))
      pos += usetc(d+pos, c);

   usetc(d+pos, 0);

   return dest;
}



/* ustrncmp:
 *  Unicode-aware version of the ANSI strncmp() function. The n parameter
 *  is in characters, on the assumption that you want to use this function 
 *  to compare only a few letters from a string.
 */
int ustrncmp(char *s1, char *s2, int n)
{
   int c1, c2;

   if (n <= 0)
      return 0;

   for (;;) {
      c1 = ugetx(&s1);
      c2 = ugetx(&s2);

      if (c1 != c2)
	 return c1 - c2;

      if ((!c1) || (--n <= 0))
	 return 0;
   }
}



/* ustricmp:
 *  Unicode-aware version of the ANSI stricmp() function.
 */
int ustricmp(char *s1, char *s2)
{
   int c1, c2;

   for (;;) {
      c1 = utolower(ugetx(&s1));
      c2 = utolower(ugetx(&s2));

      if (c1 != c2)
	 return c1 - c2;

      if (!c1)
	 return 0;
   }
}



/* ustrlwr:
 *  Unicode-aware version of the ANSI strlwr() function.
 */
char *ustrlwr(char *s)
{
   int pos = 0;
   int c, lc;

   while ((c = ugetc(s+pos)) != 0) {
      lc = utolower(c);

      if (lc != c)
	 usetat(s+pos, 0, lc);

      pos += uwidth(s+pos);
   }

   return s;
}



/* ustrupr:
 *  Unicode-aware version of the ANSI strupr() function.
 */
char *ustrupr(char *s)
{
   int pos = 0;
   int c, uc;

   while ((c = ugetc(s+pos)) != 0) {
      uc = utoupper(c);

      if (uc != c)
	 usetat(s+pos, 0, uc);

      s += uwidth(s+pos);
   }

   return s;
}



/* ustrchr:
 *  Unicode-aware version of the ANSI strchr() function.
 */
char *ustrchr(char *s, int c)
{
   int d;

   while ((d = ugetc(s)) != 0) {
      if (c == d)
	 return s;

      s += uwidth(s);
   }

   if (!c)
      return s;

   return NULL;
}



/* ustrstr:
 *  Unicode-aware version of the ANSI strstr() function.
 */
char *ustrstr(char *s1, char *s2)
{
   int len = ustrlen(s2);

   while (ugetc(s1)) {
      if (ustrncmp(s1, s2, len) == 0)
	 return s1;

      s1 += uwidth(s1);
   }

   return NULL;
}



/* ustrpbrk:
 *  Unicode-aware version of the ANSI strpbrk() function.
 */
char *ustrpbrk(char *s, char *set)
{
   char *setp;
   int c, d;

   while ((c = ugetc(s)) != 0) {
      setp = set;

      while ((d = ugetx(&setp)) != 0) {
	 if (c == d)
	    return s;
      }

      s += uwidth(s);
   }

   return NULL;
}



/* ustrtok:
 *  Unicode-aware version of the ANSI strtok() function.
 */
char *ustrtok(char *s, char *set)
{
   static char *last = NULL;
   char *prev_str, *tok, *setp;
   int c, sc;

   if (!s) {
      s = last;

      if (!s)
	 return NULL;
   }

   skip_leading_delimiters:

   prev_str = s;
   c = ugetx(&s);

   setp = set;

   while ((sc = ugetx(&setp)) != 0) {
      if (c == sc)
	 goto skip_leading_delimiters;
   }

   if (!c) {
      last = NULL;
      return NULL;
   }

   tok = prev_str;

   for (;;) {
      prev_str = s;
      c = ugetx(&s);

      setp = set;

      do {
	 sc = ugetx(&setp);
	 if (sc == c) {
	    if (!c) {
	       last = NULL;
	       return tok;
	    }
	    else {
	       s += usetat(prev_str, 0, 0);
	       last = s;
	       return tok;
	    }
	 }
      } while (sc);
   }
}



/* uatof:
 *  Unicode-aware version of the ANSI atof() function. No need to bother
 *  implementing this ourselves, since all floating point numbers are
 *  valid ASCII in any case.
 */
double uatof(char *s)
{
   return atof(uconvert(s, U_CURRENT, U_ASCII));
}



/* ustrtol:
 *  Unicode-aware version of the ANSI strtol() function. Note the
 *  nicely bodged implementation :-)
 */
long ustrtol(char *s, char **endp, int base)
{
   char *tmp = uconvert(s, U_CURRENT, U_ASCII);
   char *myendp;
   long ret;

   ret = strtol(tmp, &myendp, base);

   if (endp)
      *endp = s + uoffset(s, (long)myendp - (long)tmp);

   return ret;
}



/* information about the current format conversion mode */
typedef struct SPRINT_INFO
{
   int flags;
   int field_width;
   int precision;
   int num_special;
} SPRINT_INFO;



#define SPRINT_FLAG_LEFT_JUSTIFY             1
#define SPRINT_FLAG_FORCE_PLUS_SIGN          2
#define SPRINT_FLAG_FORCE_SPACE              4
#define SPRINT_FLAG_ALTERNATE_CONVERSION     8
#define SPRINT_FLAG_PAD_ZERO                 16
#define SPRINT_FLAG_SHORT_INT                32
#define SPRINT_FLAG_LONG_INT                 64
#define SPRINT_FLAG_LONG_DOUBLE              128
#define SPRINT_FLAG_LONG_LONG                256



static va_list sprint_args;



/* va_int:
 *  Helper for reading an integer from the varargs list.
 */
static long va_int(int flags)
{
   if (flags & SPRINT_FLAG_SHORT_INT)
      return va_arg(sprint_args, signed short int);
 #ifdef LONGLONG
   else if (flags & SPRINT_FLAG_LONG_LONG)
      return va_arg(sprint_args, signed LONGLONG);
 #endif
   else if (flags & SPRINT_FLAG_LONG_INT)
      return va_arg(sprint_args, signed long int);
   else
      return va_arg(sprint_args, signed int);
}



/* va_uint:
 *  Helper for reading an unsigned integer from the varargs list.
 */
static unsigned long va_uint(int flags)
{
   if (flags & SPRINT_FLAG_SHORT_INT)
      return va_arg(sprint_args, unsigned short int);
 #ifdef LONGLONG
   else if (flags & SPRINT_FLAG_LONG_LONG)
      return va_arg(sprint_args, unsigned LONGLONG);
 #endif
   else if (flags & SPRINT_FLAG_LONG_INT)
      return va_arg(sprint_args, unsigned long int);
   else
      return va_arg(sprint_args, unsigned int);
}



/* sprint_i:
 *  Worker function for formatting integers.
 */
static int sprint_i(char **buf, unsigned long val, int precision)
{
   char tmp[16];
   int i = 0;
   int len;

   do {
      tmp[i++] = val % 10;
      val /= 10;
   } while (val);

   for (len=i; len<precision; len++)
      *buf += usetc(*buf, '0');

   while (i > 0)
      *buf += usetc(*buf, tmp[--i] + '0');

   return len;
}



/* sprint_plus_sign:
 *  Helper to add a plus sign or space in front of a number.
 */
static void sprint_plus_sign(char **buf, SPRINT_INFO *info, int *len)
{
   if (info->flags & SPRINT_FLAG_FORCE_PLUS_SIGN) {
      *buf += usetc(*buf, '+');
      (*len)++;
   }
   else if (info->flags & SPRINT_FLAG_FORCE_SPACE) {
      *buf += usetc(*buf, ' ');
      (*len)++;
   }
}



/* sprint_int:
 *  Helper for formatting a signed integer.
 */
static int sprint_int(char **buf, SPRINT_INFO *info, long val)
{
   int len = 0;

   if (val < 0) {
      val = -val;
      *buf += usetc(*buf, '-');
      len++;
   }
   else 
      sprint_plus_sign(buf, info, &len);

   info->num_special = len;

   return sprint_i(buf, val, info->precision) + len;
}



/* sprint_unsigned:
 *  Helper for formatting an unsigned integer.
 */
static int sprint_unsigned(char **buf, SPRINT_INFO *info, unsigned long val)
{
   sprint_plus_sign(buf, info, &info->num_special);

   return sprint_i(buf, val, info->precision) + info->num_special;
}



/* sprint_hex:
 *  Helper for formatting a hex integer.
 */
static int sprint_hex(char **buf, SPRINT_INFO *info, int caps, unsigned long val)
{
   static char hex_digit_caps[] = "0123456789ABCDEF";
   static char hex_digit[] = "0123456789abcdef";

   char tmp[16];
   char *table;
   int i = 0;
   int len;

   sprint_plus_sign(buf, info, &info->num_special);

   if (info->flags & SPRINT_FLAG_ALTERNATE_CONVERSION) {
      *buf += usetc(*buf, '0');
      *buf += usetc(*buf, 'x');
      info->num_special += 2;
   }

   do {
      tmp[i++] = val & 15;
      val >>= 4;
   } while (val);

   for (len=i; len<info->precision; len++)
      *buf += usetc(*buf, '0');

   if (caps)
      table = hex_digit_caps;
   else
      table = hex_digit;

   while (i > 0)
      *buf += usetc(*buf, table[(int)tmp[--i]]);

   return len + info->num_special;
}



/* sprint_octal:
 *  Helper for formatting an octal integer.
 */
static int sprint_octal(char **buf, SPRINT_INFO *info, unsigned long val)
{
   char tmp[16];
   int i = 0;
   int len;

   sprint_plus_sign(buf, info, &info->num_special);

   if (info->flags & SPRINT_FLAG_ALTERNATE_CONVERSION) {
      *buf += usetc(*buf, '0');
      info->num_special++;
   }

   do {
      tmp[i++] = val & 7;
      val >>= 3;
   } while (val);

   for (len=i; len<info->precision; len++)
      *buf += usetc(*buf, '0');

   while (i > 0)
      *buf += usetc(*buf, tmp[--i] + '0');

   return len + info->num_special;
}



/* sprint_float:
 *  Helper for formatting a float.
 */
static int sprint_float(char **buf, SPRINT_INFO *info, double val)
{
   int len = 0;
   int v, p, i;

   if (val < 0) {
      val = -val;
      *buf += usetc(*buf, '-');
      len++;
   }
   else
      sprint_plus_sign(buf, info, &len);

   info->num_special = len;

   v = (int)val;
   val -= v;

   len += sprint_i(buf, v, -1);

   if (info->precision >= 0)
      p = info->precision;
   else
      p = 8 - len + info->num_special;

   if (p > 0) {
      *buf += usetc(*buf, '.');
      len++;

      for (i=0; i<p; i++) {
	 val *= 10;

	 if (i < p-1)
	    v = (int)val;
	 else
	    v = (int)(val + 0.5);

	 if (v < 0)
	    v = 0;
	 else if (v > 9)
	    v = 9;

	 val -= v;

	 *buf += usetc(*buf, v + '0');
	 len++;
      }
   }

   return len;
}



/* sprint_string:
 *  Helper for formatting a string.
 */
static int sprint_string(char **buf, SPRINT_INFO *info, char *s)
{
   int len = 0;
   int c;

   while ((c = ugetx(&s)) != 0) {
      if ((info->precision >= 0) && (len >= info->precision))
	 break;

      *buf += usetc(*buf, c);
      len++;
   }

   return len;
}



/* sprint:
 *  Worker function for processing a format specifier.
 */
static int sprint(char **buf, char **format, int str_pos)
{
   SPRINT_INFO info;
   char *orig = *buf;
   int *pstr_pos;
   int done, len, c, i;
   int shift, shiftbytes, shiftfiller;

   c = ugetc(*format);

   #define NEXT_C()                          \
   {                                         \
      *format += uwidth(*format);            \
      c = ugetc(*format);                    \
   }

   /* set default conversion flags */
   info.flags = 0;
   info.field_width = 0;
   info.precision = -1;
   info.num_special = 0;

   /* check for conversion flags */
   done = FALSE;

   do {
      switch (c) {

	 case '-':
	    info.flags |= SPRINT_FLAG_LEFT_JUSTIFY;
	    NEXT_C();
	    break;

	 case '+':
	    info.flags |= SPRINT_FLAG_FORCE_PLUS_SIGN;
	    NEXT_C();
	    break;

	 case ' ':
	    info.flags |= SPRINT_FLAG_FORCE_SPACE;
	    NEXT_C();
	    break;

	 case '#':
	    info.flags |= SPRINT_FLAG_ALTERNATE_CONVERSION;
	    NEXT_C();
	    break;

	 case '0':
	    info.flags |= SPRINT_FLAG_PAD_ZERO;
	    NEXT_C();
	    break;

	 default:
	    done = TRUE;
	    break;
      }

   } while (!done);

   /* check for a field width specifier */
   if (c == '*') {
      NEXT_C();
      info.field_width = va_arg(sprint_args, int);
      if (info.field_width < 0) {
	 info.flags |= SPRINT_FLAG_LEFT_JUSTIFY;
	 info.field_width = -info.field_width;
      }
   }
   else if ((c >= '0') && (c <= '9')) {
      info.field_width = 0;
      do {
	 info.field_width *= 10;
	 info.field_width += c - '0';
	 NEXT_C();
      } while ((c >= '0') && (c <= '9'));
   }

   /* check for a precision specifier */
   if (c == '.')
      NEXT_C();

   if (c == '*') {
      NEXT_C();
      info.precision = va_arg(sprint_args, int);
      if (info.precision < 0)
	 info.precision = 0;
   }
   else if ((c >= '0') && (c <= '9')) {
      info.precision = 0;
      do {
	 info.precision *= 10;
	 info.precision += c - '0';
	 NEXT_C();
      } while ((c >= '0') && (c <= '9'));
   }

   /* check for size qualifiers */
   done = FALSE;

   do {
      switch (c) {

	 case 'h':
	    info.flags |= SPRINT_FLAG_SHORT_INT;
	    NEXT_C();
	    break;

	 case 'l':
	    if (info.flags & SPRINT_FLAG_LONG_INT)
	       info.flags |= SPRINT_FLAG_LONG_LONG;
	    else
	       info.flags |= SPRINT_FLAG_LONG_INT;
	    NEXT_C();
	    break;

	 case 'L':
	    info.flags |= (SPRINT_FLAG_LONG_DOUBLE | SPRINT_FLAG_LONG_LONG);
	    NEXT_C();
	    break;

	 default:
	    done = TRUE;
	    break;
      }

   } while (!done);

   /* format the data */
   switch (c) {

      case 'c':
	 /* character */
	 *buf += usetc(*buf, va_arg(sprint_args, int));
	 len = 1;
	 NEXT_C();
	 break;

      case 'd':
      case 'i':
	 /* signed integer */
	 len = sprint_int(buf, &info, va_int(info.flags));
	 NEXT_C();
	 break;

      case 'D':
	 /* signed long integer */
	 len = sprint_int(buf, &info, va_int(info.flags | SPRINT_FLAG_LONG_INT));
	 NEXT_C();
	 break;

      case 'e':
      case 'E':
      case 'f':
      case 'g':
      case 'G':
	 /* double */
	 if (info.flags & SPRINT_FLAG_LONG_DOUBLE)
	    len = sprint_float(buf, &info, va_arg(sprint_args, long double));
	 else
	    len = sprint_float(buf, &info, va_arg(sprint_args, double));
	 NEXT_C();
	 break;

      case 'n':
	 /* store current string position */
	 pstr_pos = va_arg(sprint_args, int *);
	 *pstr_pos = str_pos;
	 len = 0;
	 NEXT_C();
	 break;

      case 'o':
	 /* unsigned octal integer */
	 len = sprint_octal(buf, &info, va_uint(info.flags));
	 NEXT_C();
	 break;

      case 'p':
	 /* pointer */
	 len = sprint_hex(buf, &info, FALSE, (unsigned long)(va_arg(sprint_args, void *)));
	 NEXT_C();
	 break;

      case 's':
	 /* string */
	 len = sprint_string(buf, &info, va_arg(sprint_args, char *));
	 NEXT_C();
	 break;

      case 'u':
	 /* unsigned integer */
	 len = sprint_unsigned(buf, &info, va_uint(info.flags));
	 NEXT_C();
	 break;

      case 'U':
	 /* unsigned long integer */
	 len = sprint_unsigned(buf, &info, va_uint(info.flags | SPRINT_FLAG_LONG_INT));
	 NEXT_C();
	 break;

      case 'x':
      case 'X':
	 /* unsigned hex integer */
	 len = sprint_hex(buf, &info, (c == 'X'), va_uint(info.flags));
	 NEXT_C();
	 break;

      default:
	 /* weird shit... */
	 len = 0;
	 break;
   }

   if (len < info.field_width) {
      if (info.flags & SPRINT_FLAG_LEFT_JUSTIFY) {
	 /* left align the result */
	 while (len < info.field_width) {
	    *buf += usetc(*buf, ' ');
	    len++;
	 }
      }
      else {
	 /* right align the result */
	 shift = info.field_width - len;

	 if (shift > 0) {
	    if (info.flags & SPRINT_FLAG_PAD_ZERO) {
	       shiftfiller = '0';

	       for (i=0; i<info.num_special; i++)
		  orig += uwidth(orig);
	    }
	    else
	       shiftfiller = ' ';

	    shiftbytes = shift * ucwidth(shiftfiller);

	    memmove(orig+shiftbytes, orig, *buf-orig);

	    for (i=0; i<shift; i++)
	       orig += usetc(orig, shiftfiller);

	    *buf += shiftbytes;
	    len += shift;
	 }
      }
   }

   return len;
}



/* uvsprintf:
 *  Unicode-aware version of the ANSI vsprintf() function.
 */
int uvsprintf(char *buf, char *format, va_list args)
{
   int len = 0;
   int c;

   memcpy(&sprint_args, &args, sizeof(sprint_args));

   while ((c = ugetx(&format)) != 0) {

      if (c == '%') {
	 if (ugetc(format) == '%') {
	    /* percent sign escape */
	    format += uwidth(format);
	    buf += usetc(buf, '%');
	    len++;
	 }
	 else {
	    /* format specifier */
	    len += sprint(&buf, &format, len);
	 }
      }
      else {
	 /* normal character */
	 buf += usetc(buf, c);
	 len++;
      }
   }

   usetc(buf, 0);

   return len;
}



/* usprintf:
 *  Unicode-aware version of the ANSI sprintf() function.
 */
int usprintf(char *buf, char *format, ...)
{
   int ret;

   va_list ap;
   va_start(ap, format);
   ret = uvsprintf(buf, format, ap);
   va_end(ap);

   return ret;
}

